from pathlib import Path
from typing import Awaitable, Callable

import aiosqlite


async def add_template_system_to_chat(db_path: Path) -> None:
    async with aiosqlite.connect(db_path) as connection:
        await connection.executescript(
            """
            ALTER TABLE chat ADD COLUMN template TEXT;
            ALTER TABLE chat ADD COLUMN system TEXT;
            """
        )


upgrades: list[tuple[str, list[Callable[[Path], Awaitable[None]]]]] = [
    ("0.1.6", [add_template_system_to_chat])
]



from textual import on
from textual.app import ComposeResult
from textual.containers import Container
from textual.screen import ModalScreen
from textual.widgets import Input, Label


class ChatRename(ModalScreen[str]):
    old_name: str = ""
    BINDINGS = [
        ("escape", "cancel", "Cancel"),
    ]

    def action_cancel(self) -> None:
        self.dismiss()

    @on(Input.Submitted)
    async def on_submit(self, event: Input.Submitted) -> None:
        if event.value:
            self.dismiss(event.value)

    def compose(self) -> ComposeResult:
        with Container(id="chat-rename-container"):
            yield Label("Rename chat", classes="title")
            yield Input(id="chat-rename-input", value=self.old_name)



import json
import os
from typing import Union, get_type_hints

from dotenv import load_dotenv

load_dotenv()


class AppConfigError(Exception):
    pass


def _parse_bool(val: Union[str, bool]) -> bool:
    return val if isinstance(val, bool) else val.lower() in ["true", "yes", "1"]


class AppConfig:
    """
    Map environment variables to class fields according to these rules:
      - Field won't be parsed unless it has a type annotation
      - Field will be skipped if not in all caps
    """

    ENV: str = "development"
    OLLAMA_URL: str = "http://localhost:11434/api"
    OTERM_VERIFY_SSL: bool = True

    def __init__(self, env):
        for field in self.__annotations__:
            if not field.isupper():
                continue

            # Raise AppConfigError if required field not supplied
            default_value = getattr(self, field, None)
            if default_value is None and env.get(field) is None:
                raise AppConfigError("The {} field is required".format(field))

            # Cast env var value to expected type and raise AppConfigError on failure
            try:
                var_type = get_type_hints(AppConfig)[field]
                if var_type == bool:
                    value = _parse_bool(env.get(field, default_value))
                elif var_type == list[str]:
                    value = env.get(field)
                    if value is None:
                        value = default_value
                    else:
                        value = json.loads(value)
                else:
                    value = var_type(env.get(field, default_value))
                self.__setattr__(field, value)
            except ValueError:
                raise AppConfigError(
                    'Unable to cast value of "{}" to type "{}" for "{}" field'.format(
                        env[field], var_type, field  # type: ignore
                    )
                )

    def __repr__(self):
        return str(self.__dict__)


# Expose Config object for app to import
Config = AppConfig(os.environ)



import json
from ast import literal_eval

from rich.text import Text
from textual.app import ComposeResult
from textual.containers import Container, Horizontal, Vertical
from textual.css.query import NoMatches
from textual.reactive import reactive
from textual.screen import ModalScreen
from textual.widgets import Button, Checkbox, Label, OptionList, Pretty, TextArea

from oterm.ollama import OllamaAPI


class ModelSelection(ModalScreen[str]):
    api = OllamaAPI()
    models = []
    models_info: dict[str, dict] = {}

    name: reactive[str] = reactive("")
    tag: reactive[str] = reactive("")
    bytes: reactive[int] = reactive(0)
    model_info: reactive[dict[str, str]] = reactive({}, layout=True)
    template: reactive[str] = reactive("")
    system: reactive[str] = reactive("")
    params: reactive[list[tuple[str, str]]] = reactive([], layout=True)
    json_format: reactive[bool] = reactive(False)

    BINDINGS = [
        ("escape", "cancel", "Cancel"),
    ]

    def action_cancel(self) -> None:
        self.dismiss()

    async def on_mount(self) -> None:
        self.models = await self.api.get_models()
        models = [model["name"] for model in self.models]
        for model in models:
            info = await self.api.get_model_info(model)
            for key in ["modelfile", "license"]:
                if key in info.keys():
                    del info[key]
            self.models_info[model] = info
        option_list = self.query_one("#model-select", OptionList)
        option_list.clear_options()
        for model in models:
            option_list.add_option(item=self.model_option(model))

    def on_option_list_option_highlighted(
        self, option: OptionList.OptionHighlighted
    ) -> None:
        model = option.option.prompt
        model_meta = next((m for m in self.models if m["name"] == str(model)), None)
        if model_meta:
            name, tag = model_meta["name"].split(":")
            self.name = name
            self.tag = tag
            self.bytes = model_meta["size"]

            self.model_info = self.models_info[model_meta["name"]]

        # Now that there is a model selected we can create the chat.
        create_button = self.query_one("#create-btn", Button)
        create_button.disabled = False

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.name == "create":
            model = f"{self.name}:{self.tag}"
            template = self.query_one(".template", TextArea).text
            template = (
                template if template != self.model_info.get("template", "") else None
            )
            system = self.query_one(".system", TextArea).text
            system = system if system != self.model_info.get("system", "") else None
            jsn = self.query_one(".json-format", Checkbox).value

            result = json.dumps(
                {
                    "name": model,
                    "template": template,
                    "system": system,
                    "format": "json" if jsn else None,
                }
            )
            self.dismiss(result)
        else:
            self.dismiss()

    @staticmethod
    def model_option(model: str) -> Text:
        return Text(model)

    def watch_name(self, name: str) -> None:
        try:
            widget = self.query_one(".name", Label)
            widget.update(f"Name: {self.name}")
        except NoMatches:
            pass

    def watch_tag(self, tag: str) -> None:
        try:
            widget = self.query_one(".tag", Label)
            widget.update(f"Tag: {self.tag}")
        except NoMatches:
            pass

    def watch_bytes(self, size: int) -> None:
        try:
            widget = self.query_one(".size", Label)
            widget.update(f"Size: {(self.bytes / 1.0e9):.2f} GB")
        except NoMatches:
            pass

    def watch_model_info(self, model_info: dict[str, str]) -> None:
        self.template = model_info.get("template", "")
        self.system = model_info.get("system", "")
        params = model_info.get("parameters", "")
        lines = params.split("\n")
        params = []
        for line in lines:
            if line:
                key, value = line.split(maxsplit=1)
                try:
                    value = literal_eval(value)
                except (SyntaxError, ValueError):
                    pass
                params.append((key, value))
        self.params = params

        try:
            widget = self.query_one(".parameters", Pretty)
            widget.update(self.params)
            widget = self.query_one(".template", TextArea)
            widget.clear()
            widget.load_text(self.template)
            widget = self.query_one(".system", TextArea)
            widget.load_text(self.system)
        except NoMatches:
            pass

    def compose(self) -> ComposeResult:
        with Container(id="model-select-container"):
            yield Label("Select a model:", classes="title")
            with Horizontal():
                with Vertical():
                    yield OptionList(id="model-select")
                    with Vertical(id="model-details"):
                        yield Label("Model info:", classes="title")
                        yield Label("", classes="name")
                        yield Label("", classes="tag")
                        yield Label("", classes="size")
                with Vertical():
                    yield Label("Template:", classes="title")
                    yield TextArea(classes="template log")
                    yield Label("System:", classes="title")
                    yield TextArea("", classes="system log")
                    yield Label("Parameters:", classes="title")
                    yield Pretty("", classes="parameters")
                    yield Label("Format", classes="title")
                    yield Checkbox("JSON output", value=False, classes="json-format")

            with Horizontal(classes="button-container"):
                yield Button(
                    "Create",
                    id="create-btn",
                    name="create",
                    disabled=True,
                    variant="primary",
                )
                yield Button("Cancel", name="cancel")



import json

from textual.app import App, ComposeResult
from textual.widgets import Footer, Header, TabbedContent, TabPane

from oterm.app.chat import ChatContainer
from oterm.app.chat_rename import ChatRename
from oterm.app.model_selection import ModelSelection
from oterm.app.splash import SplashScreen
from oterm.store.store import Store


class OTerm(App):
    TITLE = "oTerm"
    SUB_TITLE = "A terminal-based Ollama client."
    CSS_PATH = "oterm.tcss"
    BINDINGS = [
        ("n", "new_chat", "new chat"),
        ("d", "toggle_dark", "Toggle dark mode"),
        ("r", "rename_chat", "rename chat"),
        ("x", "forget_chat", "forget chat"),
        ("q", "quit", "Quit"),
    ]
    SCREENS = {
        "splash": SplashScreen(),
        "model_selection": ModelSelection(),
    }

    def action_toggle_dark(self) -> None:
        self.dark = not self.dark

    def action_quit(self) -> None:
        return self.exit()

    def action_new_chat(self) -> None:
        async def on_model_select(model_info: str) -> None:
            model: dict = json.loads(model_info)
            tabs = self.query_one(TabbedContent)
            tab_count = tabs.tab_count
            name = f"chat #{tab_count+1} - {model['name']}"
            id = await self.store.save_chat(
                id=None,
                name=name,
                model=model["name"],
                context="[]",
                template=model["template"],
                system=model["system"],
                format=model["format"],
            )
            pane = TabPane(name, id=f"chat-{id}")
            pane.compose_add_child(
                ChatContainer(
                    db_id=id,
                    chat_name=name,
                    model=model["name"],
                    system=model["system"],
                    template=model["template"],
                    format=model["format"],
                )
            )
            tabs.add_pane(pane)
            tabs.active = f"chat-{id}"

        self.push_screen("model_selection", on_model_select)

    async def action_rename_chat(self) -> None:
        tabs = self.query_one(TabbedContent)
        id = int(tabs.active.split("-")[1])
        chat = await self.store.get_chat(id)
        if chat is None:
            return
        _, name, model, context, template, system = chat

        async def on_chat_rename(name: str) -> None:
            await self.store.rename_chat(id, name)
            messages = await self.store.get_messages(id)
            tabs.remove_pane(tabs.active)
            pane = TabPane(name, id=f"chat-{id}")
            pane.compose_add_child(
                ChatContainer(
                    db_id=id,
                    chat_name=name,
                    model=model,
                    messages=messages,
                    context=context,
                    template=template,
                    system=system,
                )
            )
            tabs.add_pane(pane)
            tabs.active = f"chat-{id}"

        if chat:
            screen = ChatRename()
            screen.old_name = name
            self.push_screen(screen, on_chat_rename)

    async def action_forget_chat(self) -> None:
        tabs = self.query_one(TabbedContent)
        active_pane_id = tabs.active

        if active_pane_id:
            await self.store.delete_chat(int(active_pane_id.split("-")[1]))
            tabs.remove_pane(active_pane_id)

    async def on_mount(self) -> None:
        self.store = await Store.create()
        saved_chats = await self.store.get_chats()  # type: ignore
        if not saved_chats:
            self.action_new_chat()
        else:
            tabs = self.query_one(TabbedContent)
            for id, name, model, context, template, system, format in saved_chats:
                messages = await self.store.get_messages(id)
                pane = TabPane(name, id=f"chat-{id}")
                pane.compose_add_child(
                    ChatContainer(
                        db_id=id,
                        chat_name=name,
                        model=model,
                        context=context,
                        messages=messages,  # type: ignore
                        template=template,
                        system=system,
                        format=format,
                    )
                )
                tabs.add_pane(pane)
                tabs.active = f"chat-{id}"

        await self.push_screen("splash")

    def compose(self) -> ComposeResult:
        yield Header()
        yield TabbedContent(id="tabs")
        yield Footer()


app = OTerm()



import json
from typing import Any, AsyncGenerator, Literal

import httpx

from oterm.config import Config


class OllamaError(Exception):
    pass


class OllamaLLM:
    def __init__(
        self,
        model="nous-hermes:13b",
        template: str | None = None,
        system: str | None = None,
        context: list[int] = [],
        format: Literal["json"] | None = None,
    ):
        self.model = model
        self.template = template
        self.system = system
        self.context = context
        self.format = format

    async def completion(self, prompt: str) -> str:
        response = ""
        context = []
        async for text, ctx in self._agenerate(
            prompt=prompt,
            context=self.context,
        ):
            response = text
            context = ctx
        self.context = context
        return response

    async def stream(self, prompt) -> AsyncGenerator[str, Any]:
        context = []

        async for text, ctx in self._agenerate(
            prompt=prompt,
            context=self.context,
        ):
            context = ctx
            yield text

        self.context = context

    async def _agenerate(
        self,
        prompt: str,
        context: list[int],
    ) -> AsyncGenerator[tuple[str, list[int]], Any]:
        client = httpx.AsyncClient(verify=Config.OTERM_VERIFY_SSL)
        jsn = {
            "model": self.model,
            "prompt": prompt,
            "context": context,
        }
        if self.system:
            jsn["system"] = self.system
        if self.template:
            jsn["template"] = self.template
        if self.format:
            jsn["format"] = self.format
        res = ""

        async with client.stream(
            "POST", f"{Config.OLLAMA_URL}/generate", json=jsn, timeout=None
        ) as response:
            async for line in response.aiter_lines():
                body = json.loads(line)
                res += body.get("response", "")
                yield res, []
                if "error" in body:
                    raise OllamaError(body["error"])

                if body.get("done", False):
                    yield res, body["context"]


class OllamaAPI:
    async def get_models(self) -> list[dict[str, Any]]:
        client = httpx.AsyncClient(verify=Config.OTERM_VERIFY_SSL)
        response = await client.get(f"{Config.OLLAMA_URL}/tags")
        return response.json().get("models", []) or []

    async def get_model_info(self, model: str) -> dict[str, Any]:
        client = httpx.AsyncClient(verify=Config.OTERM_VERIFY_SSL)
        response = await client.post(f"{Config.OLLAMA_URL}/show", json={"name": model})
        if response.json().get("error"):
            raise OllamaError(response.json()["error"])
        return response.json()

    async def pull_model(self, model: str) -> None:
        client = httpx.AsyncClient()
        async with client.stream(
            "POST", f"{Config.OLLAMA_URL}/pull", json={"name": model}, timeout=None
        ) as response:
            async for line in response.aiter_lines():
                body = json.loads(line)
                if "error" in body:
                    raise OllamaError(body["error"])
                if body.get("status", "") == "success":
                    return



import aiosql

create_sqlite = """
-- name: create_chat_table
CREATE TABLE IF NOT EXISTS "chat" (
	"id"		INTEGER,
	"name"		TEXT,
	"model"		TEXT NOT NULL,
	"context"	TEXT NOT NULL,
	"template"	TEXT,
	"system"	TEXT,
	"format"	TEXT,
	PRIMARY KEY("id" AUTOINCREMENT)
);

-- name: create_message_table
CREATE TABLE IF NOT EXISTS "message" (
	"chat_id"	INTEGER NOT NULL,
	"author"	TEXT NOT NULL,
	"text"		TEXT NOT NULL,
	FOREIGN KEY("chat_id") REFERENCES "chat"("id") ON DELETE CASCADE
);

-- name: get_user_version
PRAGMA user_version;
"""

queries = aiosql.from_str(create_sqlite, "aiosqlite")



from dataclasses import dataclass
from typing import cast

from textual import events, on
from textual.app import ComposeResult
from textual.containers import Horizontal
from textual.css.query import NoMatches
from textual.message import Message
from textual.reactive import reactive
from textual.widget import Widget
from textual.widgets import Button, Input, TextArea


class PastableInput(Input):
    def _on_paste(self, event: events.Paste) -> None:
        if event.text:
            self.insert_text_at_cursor(event.text)
            if len(event.text.splitlines()) > 1:
                input = cast(FlexibleInput, self.parent.parent)  # type: ignore
                input.text = self.value
                input.toggle_multiline()
        event.stop()
        event.prevent_default()


class FlexibleInput(Widget):
    is_multiline = reactive(False)
    text = reactive("")

    @dataclass
    class Submitted(Message):
        input: "FlexibleInput"
        value: str

        @property
        def control(self) -> "FlexibleInput":
            return self.input

    def __init__(self, text, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.text = text

    def on_mount(self) -> None:
        input = self.query_one("#promptInput", PastableInput)
        textarea = self.query_one("#promptArea", TextArea)
        textarea.show_line_numbers = False
        input.focus()

    def clear(self) -> None:
        self.text = ""
        self.query_one("#promptInput", PastableInput).value = ""
        self.query_one("#promptArea", TextArea).text = ""

    def focus(self) -> "FlexibleInput":
        if self.is_multiline:
            self.query_one("#promptArea", TextArea).focus()
        else:
            self.query_one("#promptInput", PastableInput).focus()
        return self

    def toggle_multiline(self):
        self.is_multiline = not self.is_multiline
        input = self.query_one("#promptInput", PastableInput)
        textarea = self.query_one("#promptArea", TextArea)
        if self.is_multiline:
            textarea.text = self.text
            self.add_class("multiline")
            self.remove_class("singleline")
            line_count = textarea.document.line_count
            textarea.cursor_location = (
                line_count - 1,
                len(textarea.document.get_line(line_count - 1)),
            )
        else:
            input.value = self.text
            self.add_class("singleline")
            self.remove_class("multiline")
        self.focus()

    def watch_text(self):
        try:
            if len(self.text.splitlines()) > 1:
                self.query_one("#toggle-multiline", Button).disabled = True
            else:
                self.query_one("#toggle-multiline", Button).disabled = False
        except NoMatches:
            pass

    @on(PastableInput.Submitted, "#promptInput")
    def on_input_submitted(self, event: PastableInput.Submitted):
        self.post_message(self.Submitted(self, event.input.value))
        event.stop()
        event.prevent_default()

    @on(Button.Pressed, "#toggle-multiline")
    def on_toggle_multiline_pressed(self):
        self.toggle_multiline()

    @on(PastableInput.Changed, "#promptInput")
    def on_input_changed(self, event: PastableInput.Changed):
        self.text = event.input.value

    @on(TextArea.Changed, "#promptArea")
    def on_area_changed(self, event: TextArea.Changed):
        lines = [
            event.text_area.document.get_line(line)
            for line in range(event.text_area.document.line_count)
        ]

        self.text = "\n".join(lines)

    @on(Button.Pressed, "#post")
    async def on_post(self):
        self.post_message(self.Submitted(self, self.text))

    def compose(self) -> ComposeResult:
        with Horizontal():
            yield PastableInput(
                id="promptInput",
                placeholder="Message Ollama…",
            )
            yield TextArea(id="promptArea")
            with Horizontal(id="button-container"):
                yield Button("post", id="post", variant="primary")
                yield Button("↕", id="toggle-multiline", variant="success")



import asyncio
from importlib import metadata

import typer

from oterm.app.oterm import app
from oterm.store.store import Store

cli = typer.Typer()


async def upgrade_db():
    await Store.create()


@cli.command()
def oterm(
    version: bool = typer.Option(None, "--version", "-v"),
    upgrade: bool = typer.Option(None, "--upgrade"),
):
    if version:
        typer.echo(f"oterm v{metadata.version('oterm')}")
        exit(0)
    if upgrade:
        asyncio.run(upgrade_db())
        exit(0)
    app.run()


if __name__ == "__main__":
    cli()
